# -*- coding: utf-8 -*-
import sys
from logging import Logger
from kazoo.client import KazooState, KazooClient
from kazoo.security import make_digest_acl


def zk_state_listener(logger: Logger):
    def _core(state: str):
        if state == KazooState.LOST:
            logger.warning(f"已断开 zookeeper")
        elif state == KazooState.SUSPENDED:
            logger.critical(f"zookeeper 连接已中断，准备退出")
            sys.exit(-1)
        else:
            logger.info(f"已连接到zookeeper")

    return _core


def zk_ensure_path(zk: KazooClient, path: str, logger: Logger):
    # zk.ensure_path(self._nodes_path) 这个在高版本没用
    if zk.exists(path=path) is None:
        logger.info(f'没有此路径，准备创建')

        name_list = path.strip('/').split('/')
        ensure_name_list = list()
        for name in name_list:
            ensure_name_list.append(name)
            tmp_path = f'/{"/".join(ensure_name_list)}/'
            if zk.exists(path=path) is None:
                zk.create(path=tmp_path)
                logger.info(f'创建了路径: {tmp_path}')


def zk_make_acl(zk: KazooClient, path: str, username: str, password: str, logger: Logger):
    zk_ensure_path(zk=zk, path=path, logger=logger)
    acl = make_digest_acl(username=username, password=password, all=True)
    zk.set_acls(path, acls=[acl, ])
    logger.info(f'[SUCCESS]ACL创建成功: {path}')



# -*- coding: utf-8 -*-
import abc
import sys
import logging
from typing import Set, List


class AbstractNodeChangeService(abc.ABC):
    logger = logging.getLogger(__name__)

    def __init__(self):
        self._last_node_id_set: Set | None = None

    def on_change(self, current_node_id_list: List[str]):
        if self._last_node_id_set is None:
            self._last_node_id_set = set(current_node_id_list)
            if len(current_node_id_list) > 0:
                self.logger.critical(f"服务端启动时，节点列表却不为空，无法继续")
                sys.exit(-1)
            return

        new_node_id_set = set(current_node_id_list)

        online_node_id_set = new_node_id_set - self._last_node_id_set
        offline_node_id_set = self._last_node_id_set - new_node_id_set

        self.handle(online_node_id_set, offline_node_id_set)

        # 保存last_node_id_set
        self._last_node_id_set = new_node_id_set

    @abc.abstractmethod
    def handle(self, online_node_id_set: set[str], offline_node_id_set: set[str]):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
import time
import sys
from kazoo.exceptions import NoNodeError
from kazoo.client import KazooClient
from kazoo.recipe.watchers import ChildrenWatch

from a3kazoo.utils import zk_state_listener, zk_ensure_path
from .abstract_node_change_service import AbstractNodeChangeService


class WatchNodesDAO:

    def __init__(self, conf: dict, nodes_path: str, node_change_service: AbstractNodeChangeService):
        self._conf = conf
        self._nodes_path = nodes_path
        self._node_change_service = node_change_service
        self._logger = self._node_change_service.logger

    def run(self):
        zk = KazooClient(**self._conf)
        zk.add_listener(zk_state_listener(self._logger))
        zk.start()

        try:
            zk_ensure_path(zk=zk, path=self._nodes_path, logger=self._logger)
        except NoNodeError:
            self._logger.critical(f"NoNodeError: 请确认路径与ACL, {self._nodes_path}")
            zk.stop()
            zk.close()
            sys.exit(-1)

        self._logger.info(f"开启观察...")
        ChildrenWatch(zk, self._nodes_path, self._node_change_service.on_change)
        while True:
            time.sleep(60)
